#ifndef DUNE_BLOCK_GAUSS_SEIDEL_STEP_HH
#define DUNE_BLOCK_GAUSS_SEIDEL_STEP_HH

#include <dune/common/bitsetvector.hh>

#include "lineariterationstep.hh"

/** \brief A linear Gauß-Seidel iteration step for convex problems which have a block-structure.
 * 
 *  \tparam OperatorType The linear operator type
 *  \tparam DiscFuncType The block vector type of the right hand side and the iterates
 *  \tparam BitVectorType The type of the bit-vector specifying degrees of freedom that shall be ignored.
 *
 */
template<class OperatorType, 
         class DiscFuncType,
         class BitVectorType = Dune::BitSetVector<DiscFuncType::block_type::dimension> >
         class BlockGSStep : public LinearIterationStep<OperatorType, DiscFuncType, BitVectorType>
    {

        typedef typename DiscFuncType::block_type VectorBlock;

        enum {BlockSize = VectorBlock::dimension};

    public:
        
        //! Default constructor.  Doesn't init anything
        BlockGSStep() {}

        //! Constructor with a linear problem
        BlockGSStep(const OperatorType& mat, DiscFuncType& x, const DiscFuncType& rhs)
            : LinearIterationStep<OperatorType,DiscFuncType>(mat, x, rhs)
        {}

        //! Perform one iteration
        virtual void iterate();
    
        //! Return the solution
        virtual DiscFuncType getSol();

        /** \brief Compute the residual of the current iterate of a (block) degree of freedom
         *
         *  \param index Global index of the (block) degree of freedom 
         *  \param r Write residual in this vector
         */
        virtual void residual(int index, VectorBlock& r) const;

    };


#include "blockgsstep.cc"

#endif
