#include <dune/solvers/transferoperators/truncatedmgtransfer.hh>
#include <dune/solvers/iterationsteps/projectedblockgsstep.hh>

#include <dune/solvers/computeenergy.hh>


#if HAVE_IPOPT
#include <dune/solvers/solvers/quadraticipopt.hh>
#endif

template <class MatrixType, class VectorType>
void MonotoneMGStep<MatrixType, VectorType>::
preprocess()
{
    // Call preprocess of the base class
    MultigridStep<MatrixType,VectorType>::preprocess();

    // Then specify the subset of entries to be reassembled at each iteration
    recompute_.resize(this->numLevels());
    recompute_[recompute_.size()-1] = (*hasObstacle_)[recompute_.size()-1];
    for (int i=this->mgTransfer_.size()-1; i>=0; i--)
        this->mgTransfer_[i]->restrictScalarBitField(recompute_[i+1], recompute_[i]);
        

    for (size_t i=0; i<this->mgTransfer_.size(); i++)
        dynamic_cast<TruncatedMGTransfer<VectorType>*>(this->mgTransfer_[i])->recompute_ = &recompute_[i];
    
    // /////////////////////////////////////////////
    //   Set up base solver
    // /////////////////////////////////////////////
    if (typeid(*this->basesolver_) == typeid(LoopSolver<VectorType>)) {

        LoopSolver<VectorType>* loopBaseSolver = dynamic_cast<LoopSolver<VectorType>* > (this->basesolver_);
     
        typedef ProjectedBlockGSStep<MatrixType, VectorType> SmootherType;
        
        dynamic_cast<SmootherType*>(loopBaseSolver->iterationStep_)->hasObstacle_ = &(*hasObstacle_)[0];
        dynamic_cast<SmootherType*>(loopBaseSolver->iterationStep_)->obstacles_   = &(*obstacles_)[0];
        
#if HAVE_IPOPT
    } else if (typeid(*this->basesolver_) == typeid(QuadraticIPOptSolver<MatrixType,VectorType>)) {

        QuadraticIPOptSolver<MatrixType,VectorType>* ipoptBaseSolver = dynamic_cast<QuadraticIPOptSolver<MatrixType,VectorType>*> (this->basesolver_);
     
        ipoptBaseSolver->obstacles_ = &(*obstacles_)[0];
#endif
    } else {
        DUNE_THROW(SolverError, "You can't use " << typeid(*this->basesolver_).name()
                   << " as a base solver for contact problems!");
    }
}

template<class MatrixType, class VectorType>
void MonotoneMGStep<MatrixType, VectorType>::nestedIteration()
{
    for (int i=this->numLevels()-1; i>0; i--) {

        Dune::BitSetVector<dim> dummy(this->rhsHierarchy_[i].size(), false);

        // /////////////////////////////
        //   Restrict obstacles
        // //////////////////////////////
        obstacleRestrictor_->restrict((*obstacles_)[i], 
                                    (*obstacles_)[i-1],
                                    (*hasObstacle_)[i], 
                                    (*hasObstacle_)[i-1],
                                    *this->mgTransfer_[i-1],
                                    dummy);

        // Restrict right hand side
        this->mgTransfer_[i-1]->restrict(this->rhsHierarchy_[i], this->rhsHierarchy_[i-1]);
    }

    for (this->level_ = 0; this->level_<(int) this->numLevels()-1; this->level_++) {

        // If we start from an infeasible configuration, the restricted
        // obstacles may be inconsistent.  We do an ad hoc correction here.
        // The true obstacles on the maxlevel are not touched.
        for (size_t i=0; i<(*obstacles_)[this->level_].size(); i++) {
            BoxConstraint<field_type,dim>& cO = (*obstacles_)[this->level_][i];
            for (int j=0; j<dim; j++)
                if (cO.lower(j) > cO.upper(j))
                    cO.lower(j) = cO.upper(j) = (cO.lower(j) + cO.upper(j)) / 2;
        }
        
        std::cout << "Nested iteration on level " << this->level_ << std::endl;
        iterate();
        iterate();
        //std::cout << this->xHierarchy_[this->level_] << std::endl;
        this->mgTransfer_[this->level_]->prolong(*this->xHierarchy_[this->level_], *this->xHierarchy_[this->level_+1]);
        
    }
     
}


template <class MatrixType, class VectorType>
void MonotoneMGStep<MatrixType, VectorType>::iterate()
{
    this->preprocessCalled = false;

    int& level = this->level_;

    // Define references just for ease of notation
    std::vector<Dune::shared_ptr<const MatrixType> >& mat = this->matrixHierarchy_;
    std::vector<Dune::shared_ptr<VectorType> >& x   = this->xHierarchy_;
    std::vector<VectorType>& rhs = this->rhsHierarchy_;
    std::vector<std::vector<BoxConstraint<field_type,dim> > >& obstacles = *this->obstacles_;

    Dune::BitSetVector<dim> critical(x[level]->size(), false);

    // Solve directly if we're looking at the coarse problem
    if (level == 0) {

        this->basesolver_->solve();

        // Determine critical nodes  (only for debugging)
        const double eps = 1e-12;
        for (size_t i=0; i<obstacles[level].size(); i++) {
            
            for (int j=0; j<dim; j++) {
                
                if (obstacles[level][i].lower(j) >= (*x[level])[i][j] - eps
                    || obstacles[level][i].upper(j) <= (*x[level])[i][j] + eps) {
                    critical[i][j] = true;
                }
                
            }
        }

    } else {
    
        // Presmoothing
        ProjectedBlockGSStep<MatrixType,VectorType>* presmoother = dynamic_cast<ProjectedBlockGSStep<MatrixType, VectorType>*>(this->presmoother_[level].get());
        assert(presmoother);
        presmoother->setProblem(*(mat[level]), *x[level], rhs[level]);
        presmoother->obstacles_ = &(obstacles[level]);
        presmoother->hasObstacle_ = &((*hasObstacle_)[level]);
        presmoother->ignoreNodes_ = this->ignoreNodesHierarchy_[level];
        
        for (int i=0; i<this->nu1_; i++)
            presmoother->iterate();
        
        // First, a backup of the obstacles for postsmoothing
        std::vector<BoxConstraint<field_type,dim> > obstacleBackup = obstacles[level];
        
        // Compute defect obstacles
        for (size_t i=0; i<(*obstacles_)[level].size(); i++)
            obstacles[level][i] -= (*x[level])[i];
        
        // ///////////////////////
        //    Truncation
        // ///////////////////////
        
        // Determine critical nodes
        const double eps = 1e-12;
        for (size_t i=0; i<obstacles[level].size(); i++) {
            
            for (int j=0; j<dim; j++) {
                
                if (obstacles[level][i].lower(j) >= -eps
                    || obstacles[level][i].upper(j) <= eps) {
                    critical[i][j] = true;
                }
                
            }
        }
        
     
        // Restrict stiffness matrix
        dynamic_cast<TruncatedMGTransfer<VectorType>*>(this->mgTransfer_[level-1])
            ->galerkinRestrict(*(mat[level]), *(const_cast<MatrixType*>(mat[level-1].get())), critical);
        
        // Restrict obstacles
        obstacleRestrictor_->restrict(obstacles[level], obstacles[level-1],
                                    (*hasObstacle_)[level], (*hasObstacle_)[level-1],
                                    *this->mgTransfer_[level-1],
                                    critical);
        
        // //////////////////////////////////////////////////////////////////////
        // Restriction:  fineResiduum = rhs[level] - mat[level] * x[level];
        // //////////////////////////////////////////////////////////////////////
        VectorType fineResidual = rhs[level];
        mat[level]->mmv(*x[level], fineResidual);

        // restrict residual
        dynamic_cast<TruncatedMGTransfer<VectorType>*>(this->mgTransfer_[level-1])->restrict(fineResidual, rhs[level-1], critical);
        
        // Choose all zeros as the initial correction
        *x[level-1] = 0;
        
        // ///////////////////////////////////////
        // Recursively solve the coarser system
        // ///////////////////////////////////////
        level--;
        for (int i=0; i<this->mu_; i++)
            iterate();
        level++;
        
        // ////////////////////////////////////////
        // Prolong
        // ////////////////////////////////////////
        
        // add correction to the presmoothed solution
        VectorType tmp;
        dynamic_cast<TruncatedMGTransfer<VectorType>*>(this->mgTransfer_[level-1])->prolong(*x[level-1], tmp, critical);
        *x[level] += tmp;
        
        // restore the true (non-defect) obstacles
        obstacles[level] = obstacleBackup;
        
#ifndef NDEBUG
        // Debug: is the current iterate really admissible?
        for (size_t i=0; i<obstacles[level].size(); i++)
            for (int j=0; j<VectorType::block_type::dimension; j++)
                if ((*x[level])[i][j] < obstacles[level][i].lower(j)
                    || (*x[level])[i][j] > obstacles[level][i].upper(j)) {
                    
                    std::cout << "Obstacle disregarded!\n";
                    std::cout << (*x[level])[i] << std::endl << obstacles[level][i] << std::endl;
                    std::cout << "level: " << level << "   index: " << i << "   komponent: " << j << std::endl;
                    std::cout << "is " << ((critical[i][j]) ? "" : "not") << "critical" << std::endl;
                }
#endif
        
        // Postsmoothing
        ProjectedBlockGSStep<MatrixType,VectorType>* postsmoother = dynamic_cast<ProjectedBlockGSStep<MatrixType, VectorType>*>(this->postsmoother_[level].get());
        assert(postsmoother);
        postsmoother->setProblem(*(mat[level]), *x[level], rhs[level]);
        postsmoother->obstacles_ = &obstacles[level];
        postsmoother->hasObstacle_ = &((*hasObstacle_)[level]);
        postsmoother->ignoreNodes_ = this->ignoreNodesHierarchy_[level];
        
        for (int i=0; i<this->nu2_; i++)
            postsmoother->iterate();
        
    }
    
    // ////////////////////////////////////////////////////////////////////
    //   Track the number of critical nodes found during this iteration
    // ////////////////////////////////////////////////////////////////////
    if (level==(int) this->numLevels()-1 && this->verbosity_==NumProc::FULL) {
            
        std::cout << critical.count() << " critical nodes found on level " << level;
            
        int changes = 0;
        for (unsigned int i=0; i<oldCritical.size(); i++)
            for (int j=0; j<dim; j++)
                if (oldCritical[i][j] !=critical[i][j])
                    changes++;
        
        std::cout << ", and " << changes << " changes." << std::endl;
        oldCritical = critical;
    }

    // Debug: output energy
    if (level==(int) this->numLevels()-1 && this->verbosity_==NumProc::FULL)
        std::cout << "Total energy: " 
                  << std::setprecision(10) << computeEnergy(*mat[level], *x[level], rhs[level]) << std::endl;

}
