template<class MatrixType, class VectorType>
inline
VectorType TrustRegionGSStep<MatrixType, VectorType>::getSol()
{
    return *(this->x_);
}

template<class MatrixType, class VectorType>
inline
typename VectorType::field_type TrustRegionGSStep<MatrixType, VectorType>::
residual(unsigned int index, int blockIndex) const
{
    const MatrixType& mat = *this->mat_;

    typedef typename MatrixType::row_type RowType;
    typedef typename RowType::ConstIterator ColumnIterator;

    /* The following loop computes
     * \f[ sum_i = \sum_{j \ne i} A_{ij}x_j \f]
     */
    field_type r = (*this->rhs_)[index][blockIndex];
    ColumnIterator cIt    = mat[index].template begin();
    ColumnIterator cEndIt = mat[index].template end();

    for (; cIt!=cEndIt; ++cIt) {
        // r_i -= A_ij x_j
        // r -= (*cIt)[blockIndex] * (*this->x_)[cIt.index()];
        for (int i=0; i<blocksize; i++) {

            // Omit diagonal element
            if (cIt.index()==index && i==blockIndex)
                continue;

            r -= (*cIt)[blockIndex][i] * (*this->x_)[cIt.index()][i];

        }

    }

    //std::cout << "+++ index " << index << "  rhs: " << (*this->rhs_)[index] << std::endl;
    return r;

}

template<class MatrixType, class VectorType>
inline
void TrustRegionGSStep<MatrixType, VectorType>::iterate()
{
    const MatrixType& mat = *this->mat_;
    const std::vector<BoxConstraint<field_type,blocksize> >& obstacles = *this->obstacles_;

    // Debug: check for energy decrease
    //std::cout << "----------------------------------\n";
    //double oldEnergy = computeEnergy(mat, *this->x_, *this->rhs_);

    for (size_t i=0; i<this->x_->size(); i++) {

        // scalar gauss-seidel
        for (int j=0; j<blocksize; j++) {

            if (this->ignoreNodes_ and (*this->ignoreNodes_)[i][j])
                continue;

            // Compute residual
            field_type r = residual(i, j);
            const field_type& diag = mat[i][i][j][j];

            // Find local minimum
            field_type& x = (*this->x_)[i][j];

            if (diag > 0) {
                //printf("%d %d konvex\n", i, j);
                //double oldEnergy = computeEnergy(mat, *this->x_, *this->rhs_);
                // 1d problem is convex
#ifndef NDEBUG
                field_type oldX = x;
#endif
                x = r / diag;
#ifndef NDEBUG
                field_type energyX = 0.5*diag*x*x - r*x;
                field_type energyOldX = 0.5*diag*oldX*oldX - r*oldX;
                assert(energyX <= energyOldX + 1e-6);
#endif
                if ( x < obstacles[i].lower(j)) {
                    // assert that projection increases functional value
//                     double trueMin = 0.5*diag*x*x - r*x;
//                     double proMin  = 0.5*diag*obstacles[i].lower(j)*obstacles[i].lower(j) - r*obstacles[i].lower(j);
                    //assert(trueMin < proMin);
                    x = obstacles[i].lower(j);
                }
                else if ( x > obstacles[i].upper(j)) {
                    // assert that projection increases functional value
//                     double trueMin = 0.5*diag*x*x - r*x;
//                     double proMin  = 0.5*diag*obstacles[i].upper(j)*obstacles[i].upper(j) - r*obstacles[i].upper(j);
                    //assert(trueMin < proMin);
                    x = obstacles[i].upper(j);
                }

                // Debug: check for energy decrease
                //double energy = computeEnergy(mat, *this->x_, *this->rhs_);
                //assert(oldEnergy+1e-6>=energy);

            } else  {
                //printf("%d %d konkav\n", i, j);
                // 1d problem is concave or linear.
                // Minimum is attained at one of the boundaries
                field_type lBound = obstacles[i].lower(j);
                field_type uBound = obstacles[i].upper(j);

                field_type lValue = 0.5*diag*lBound*lBound - r*lBound;
                field_type uValue = 0.5*diag*uBound*uBound - r*uBound;

#if 0
                double max = 0.5*diag*(r/diag)*(r/diag) - r*r/diag;
                assert(max >= lValue);
                assert(max >= uValue);
#endif

                x = (lValue < uValue) ? lBound : uBound;

                // Debug: check for energy decrease
                //double energy = computeEnergy(mat, *this->x_, *this->rhs_);
                //assert(oldEnergy+1e-6>=energy);

            }

        }

    }

    // Debug: check for energy decrease
    //double energy = computeEnergy(mat, *this->x_, *this->rhs_);
    //assert(oldEnergy+1e-6>=energy);
}
